# encoding=utf-8

from itertools import chain
import warnings

from logbook import Logger

from hansken.remote import ProjectContext
from hansken.tool import add_command


log = Logger(__name__)


try:
    from fuse import FUSE, Operations
    # REG = regular file, LNK = symbolic link
    from stat import S_IFREG, S_IFLNK

    def run_mount_command(args):
        if not args.endpoint:
            mount_parser.error('the following argument is required: ENDPOINT')
        if not args.keystore:
            mount_parser.error('the following argument is required: KEYSTORE')

        with ProjectContext(args.endpoint, args.project,
                            keystore_url=args.keystore, auth=args.auth, verify=args.verify) as context:
            images = [image['id'] for image in context.images()]

            options = {
                # align the fs' name with common unix-style filesystem names
                'fsname': 'hanskenfs',
                # default the maximum read to 512KiB (runtime might force this lower)
                'max_read': 512 << 10,
            }

            # parse mount options from the -o arg as a comma-separated list of options
            # NB: FUSE will turn this back into a -o style line, but offers no way of passing this verbatim
            for option in (args.options.split(',') if args.options else []):
                option = option.split('=', 1)
                if len(option) == 2:
                    options[option[0]] = option[1]
                else:
                    # treat singular keys as flags
                    options[option[0]] = True

            operations = HanskenFS(context, *images)
            # let fuse mount the actual filesystem
            FUSE(operations, args.mount_point, foreground=args.foreground, **options)

    class HanskenFS(Operations):
        """
        A subset of filesystem operations that enable a number of images
        within a project to be accessed as read-only local files. All metadata
        calls are cached locally, read calls are passed to remote. Depending
        on mount options, the local system may cache blocks read from this
        filesystem.
        """

        def __init__(self, context, *images):
            log.debug('gathering info to mount {} images in project {}', len(images), context.project_id)
            # gather the root traces of all the images, used for filesystem metadata and reading data
            self.roots = {'/{}'.format(root.image_id): root for root in context.roots()
                          if root.image_id in images}

            log.debug('got {} roots', len(self.roots), len(images))
            # store relevant image metadata to create 'symlinks' to the image data
            self.links = {'/{}'.format(image['description']): image['id'] for image in context.images()
                          if image['id'] in images and image.get('description')}

            log.debug('got {} descriptions to be used as symlinks', len(self.links))
            # cache the keys to be used for file access (treat 404s as 'no key needed')
            self.keys = {'/{}'.format(image): context.connection.key(image, raise_on_not_found=False)
                         for image in images}

            log.debug('got {}/{} keys', len(self.keys) - list(self.keys.values()).count(None), len(self.keys))

        def _stat(self, path):
            """
            A local minimized implementation of stat/fstat, providing only
            values for ``st_size`` and ``st_mtime`` (actually birth time, but
            the ``stat`` struct seems to not include this).

            :param path: the path to stat (starting with ``/``)
            :return: the size and mtime of *path*, as (`int`, `float`)
            """
            root = self.roots.get(path)
            size = root.get('extracted.data.raw.size') or 0
            mtime = root.get('processed.origin.createdOn') or root.get('processed.tool.meta.publishedOn')
            mtime = mtime.timestamp() if mtime else 0.0

            return size, mtime

        def getattr(self, path, fh=None):
            log.debug('hanskenfs: getattr({}, fh={})', path, fh)

            if path in self.roots:
                # getattr() for a 'file', return it as a regular file with access rights u+r,g+r,o+r
                size, mtime = self._stat(path)
                return dict(  # noqa: C408
                    st_mode=S_IFREG | 0o444,
                    st_size=size,
                    st_mtime=mtime
                )

            if path in self.links:
                # getattr for a 'link', return it as a link with access rights u+rx,g+rx,o+rx
                size, mtime = self._stat('/{}'.format(self.links.get(path)))
                return dict(  # noqa: C408
                    st_mode=S_IFLNK | 0o555,
                    st_size=size,
                    st_mtime=mtime,
                    st_nlinks=1
                )

            # not something we 'understand', let fusepy handle this one
            return super(HanskenFS, self).getattr(path, fh=fh)

        def readdir(self, path, fh=None):
            log.debug('hanskenfs: readdir({}, {})', path, fh)
            # strip the leading slashes from all the paths we're tracking
            return ['.', '..'] + [f.lstrip('/') for f in chain(self.roots.keys(), self.links.keys())]

        def readlink(self, path):
            log.debug('hanskenfs: readlink({})', path)
            return self.links.get(path)

        def read(self, path, size, offset, fh=None):
            log.debug('hanskenfs: read({}, {}, {}, {})', path, size, offset, fh)
            if path in self.roots:
                # return all bytes in the buffer, closing the underlying connection for reuse
                with self.roots.get(path).open(offset=offset, size=size, key=self.keys.get(path)) as buffer:
                    return buffer.read()

            # let fuse generate a useful error
            super(HanskenFS, self).read(path, size, offset, fh)

    fusepy_available = True
except (ImportError, OSError) as e:
    # allow someone reading the debug log to see why hansken mount is unavailable
    log.debug('mount command not available', e)
    fusepy_available = False

    # define a run_mount that will indicate error to user
    def run_mount_command(*args, **kwargs):
        mount_parser.error("command requires FUSE's python bindings, install hansken.py with the 'mount' extra or "
                           "install package 'fusepy' manually")


def run_mount(*args, **kwargs):
    warnings.warn('{0}.run_mount has been renamed to {0}.run_mount_command'.format(__name__),
                  DeprecationWarning)
    run_mount_command(*args, **kwargs)


help_text = 'mount images linked to a project as files inside an empty directory'
if not fusepy_available:
    # extend help text in case we'll expect errors
    help_text += ' (requires fusepy to be installed)'

mount_parser = add_command('mount',
                           run_mount_command,  # either the real deal or an error-wrapper
                           optional_project=False,
                           help=help_text)
mount_parser.add_argument('project', metavar='PROJECT',
                          help='the project to mount images from')
mount_parser.add_argument('mount_point', metavar='MOUNT_POINT',
                          help='an empty directory to use as the mount point')
mount_parser.add_argument('-f', '--foreground', action='store_true', default=False,
                          help='keep the process in the foreground (the default is to background the process)')
mount_parser.add_argument('-o', '--options', metavar='OPTIONS', default='',
                          help='FUSE mount options (see fuse documentation), specified like the mount command')



from abc import ABC, abstractmethod

from hansken.util import DictView


class AbstractTrace(ABC, DictView):
    @abstractmethod
    def update(self, key_or_updates=None, value=None, data=None, overwrite=False):
        """
        Requests the remote to update or add metadata properties for this
        `.Trace`.

        :param key_or_updates: either a `str` (the metadata property to be
            updated) or a mapping supplying both keys and values to be updated
            (or `None` if only data is supplied)
        :param value: the value to update metadata property *key* to (used
            only when *key_or_updates* is a `str`)
        :param data: a `dict` mapping data type / stream name to bytes to be
            imported
        :param overwrite: whether properties to be imported should be
            overwritten if already present
        :return: processing information from remote
        """

    @abstractmethod
    def open(self, stream='raw', offset=0, size=None):
        """
        Open a data stream of a named stream (default ``raw``) for this
        `.Trace`.

        .. note::
            Multiple calls to `read(num_bytes)` on the stream resulting from
            this call works fine in Python 3, but will fail in Python 2.

        :param stream: stream to read
        :param offset: byte offset to start the stream on
        :param size: the number of bytes to make available
        :return: a file-like object to read bytes from the named stream
        :rtype: `io.BufferedReader`
        """

    @abstractmethod
    def child_builder(self, name=None):
        """
        Create a `.TraceBuilder` to build a trace to be saved as a child of
        this `.Trace`. Note that ``name`` is a mandatory property for a trace,
        even though it is optional here. A ``name`` can be added later using
        the `.TraceBuilder.update` method. Furthermore, a new trace will only
        be added to the index once explicitly saved (e.g. through
        `.TraceBuilder.build`).

        :param name: the name for the trace being built
        :return: a `.TraceBuilder` set up to create a child trace of this
            `.Trace`
        """



# encoding=utf-8

from datetime import datetime

from iso8601 import parse_date
from logbook import Logger
from tabulate import tabulate, tabulate_formats

from hansken.remote import Connection
from hansken.tool import add_command
from hansken.util import omit_empty


log = Logger(__name__)


RELEVANT_DATES = {
    'queued': 'queuedOn',
    'running': 'startedOn',
    'cancelled': 'endedOn',
    'failed': 'endedOn',
    'completed': 'endedOn',
}


def filter_tasks(items, **filters):
    def item_matches(item):
        position, task = item
        # task matches if there's no filters provided or all of them match
        return not filters or all(task.get(key) == value for key, value in filters.items())

    yield from filter(item_matches, items)


def format_task(position, task):
    # select relevant values from the task
    return (
        position,
        task.get('id'),
        task.get('type'),
        task.get('targetProject'),
        task.get('status'),
        # determine relevant value for "since" by status
        task.get(RELEVANT_DATES.get(task.get('status'))),
    )


def render_tasks(tasks, tablefmt='simple', **filters):
    # pull relevant information from response (position on the queue (if present) and the task itself)
    tasks = ((item.get('queuePosition'), item.get('task')) for item in tasks)
    # prepare a 6-column table
    headers = ('position', 'id', 'type', 'project', 'status', 'since')
    tasks = [format_task(position, task) for position, task in filter_tasks(tasks, **filters)]
    # sort tabular data by position and the "since" column
    tasks = sorted(tasks, key=lambda row: (row[0] if row[0] is not None else -1, row[-1]))
    return tabulate(tasks, headers=headers, tablefmt=tablefmt)


def run_tasks_command(args):
    if not args.endpoint:
        tasks_parser.error('the following argument is required: ENDPOINT')

    if args.start and not args.end:
        args.end = datetime.now(tz=args.timezone)
        log.info('defaulting end of tasks range to {}', args.end)

    with Connection(args.endpoint, auth=args.auth, verify=args.verify) as connection:
        kwargs = omit_empty({
            'state': args.state,
            'project_id': args.project,
            'start': args.start,
            'end': args.end,
        })
        log.info('using {} selectors on tasks to be listed ({})', len(kwargs), set(kwargs.keys()))

        filters = omit_empty({
            'status': args.status,
        })
        log.info('applying {} filters on tasks to be listed ({})', len(filters), set(filters.keys()))

        # select tasks from remote by kwargs, apply filters locally
        print(render_tasks(connection.tasks(**kwargs), tablefmt=args.format, **filters))


tasks_parser = add_command('tasks', run_tasks_command, optional_project=True,
                           help='list either open or closed tasks tracked by the Hansken scheduler')

tasks_parser.add_argument('--state', metavar='STATE', choices=('open', 'closed'), default='open',
                          help='select state of tasks to list, one of {} (defaults to open)'
                               ''.format(', '.join(('open', 'closed'))))
tasks_parser.add_argument('--status', metavar='STATUS',
                          help='status of the tasks to list, e.g. queued or running')
tasks_parser.add_argument('--from', dest='start', type=parse_date,
                          help='minimum relevant date (optional time) tasks to list, in ISO 8601 format')
tasks_parser.add_argument('--to', dest='end', type=parse_date,
                          help='maximum relevant date (optional time) tasks to list, in ISO 8601 format '
                               '(automatically set to now if omitted)')
tasks_parser.add_argument('-f', '--format', metavar='FORMAT', choices=tabulate_formats, default='simple',
                          help='specify tabulation format, one of {} (see documentation of python module tabulate)'
                               ''.format(', '.join(tabulate_formats)))



# encoding=utf-8

import warnings

import iso8601
from tabulate import tabulate, tabulate_formats

from hansken import __version__
from hansken.query import Term
from hansken.remote import Connection, ProjectContext
from hansken.tool import add_command
from hansken.util import format_byte_size


def run_versions_command(args):
    """
    Prints versions of both ``hansken.py`` and the remote it connects to.
    Optionally also retrieves the remote versions at extraction time for all
    images extracted for a project.

    :param args: an `argparse.Namespace` object with parsed arguments
    :return: 0
    """
    if not args.endpoint:
        versions_parser.error('the following argument is required: ENDPOINT')

    with Connection(args.endpoint, keystore_url=args.keystore, auth=args.auth, verify=args.verify) as connection:
        # get the build version, turn the build date into a proper date
        remote_versions = connection.version()
        build_date = iso8601.parse_date(remote_versions.get('timestamp')).date()

        print('hansken.py:    {}'.format(__version__))
        print('REST endpoint: {} (built on {})'.format(remote_versions.get('build'), build_date))

        if args.project:
            # leave a newline between this section and the next
            print()

            with ProjectContext(connection, args.project) as context:
                # create an info dict per image
                images = context.images()
                images = {image['id']: image for image in images}
                # gather some additional information from the root traces
                roots = context.roots()
                for root in roots:
                    version = root.tool_versions
                    # extraction service inserts its own version in the root trace
                    # named changed a few times, in reverse chronological order:
                    names = ('extraction-framework', 'hansken-extraction-service', 'extraction-service')
                    version = next(
                        # find the first non-falsy value, (…)
                        iter(filter(None, (version.get(name) for name in names))),
                        # (…) default to (unknown)
                        '(unknown)'
                    )

                    date = root.get('processed.tool.meta.extractionStartedOn')
                    date = iso8601.parse_date(date).date() if date else '(unknown)'

                    image = images.get(root.image_id)
                    # include additional info available from the root trace
                    image.update(
                        version=version,
                        date=date,
                        size=format_byte_size(root.data.raw.size)
                    )

                    if args.traces:
                        # only perform the search when needed, expensive compared to the rest
                        image.update(traces=len(context.search(Term('image', image.get('id')), count=0)))

                # predefine the headers to use
                headers = ['id', 'description', 'version', 'date']
                if args.size:
                    headers.append('size')
                if args.traces:
                    headers.append('traces')

                # reduce the data set to the bare essentials
                images = [
                    [value.get(header) for header in headers]
                    for value in sorted(images.values(), key=lambda value: value.get('id'))
                ]
                # and print the data set as a table
                print(tabulate(images, headers=headers, tablefmt=args.format))

    return 0


def run_versions(*args, **kwargs):
    warnings.warn('{0}.run_versions has been renamed to {0}.run_versions_command'.format(__name__),
                  DeprecationWarning)
    run_versions_command(*args, **kwargs)


# define command versions
versions_parser = add_command('versions',
                              run_versions_command,
                              optional_project=False,
                              help='print client, server and extraction versions')
versions_parser.add_argument('project', metavar='PROJECT', nargs='?',
                             help='print versions at extraction time for images in PROJECT')
versions_parser.add_argument('-s', '--size', dest='size', action='store_true', default=False,
                             help='display the raw data size of images')
versions_parser.add_argument('-S', '--no-size', dest='size', action='store_false',
                             help="don't display the raw data size of images (this is the default)")
versions_parser.add_argument('-t', '--traces', dest='traces', action='store_true', default=False,
                             help='display the number of traces per image')
versions_parser.add_argument('-T', '--no-traces', dest='traces', action='store_false',
                             help="don't display the number of traces per image (this is the default)")
versions_parser.add_argument('-f', '--format', metavar='FORMAT', choices=tabulate_formats, default='simple',
                             help='specify tabulation format, one of {} (see documentation of python module tabulate)'
                                  ''.format(', '.join(tabulate_formats)))
